"use strict";Object.defineProperty(exports, "__esModule", {value: true}); function _nullishCoalesce(lhs, rhsFn) { if (lhs != null) { return lhs; } else { return rhsFn(); } } function _optionalChain(ops) { let lastAccessLHS = undefined; let value = ops[0]; let i = 1; while (i < ops.length) { const op = ops[i]; const fn = ops[i + 1]; i += 2; if ((op === 'optionalAccess' || op === 'optionalCall') && value == null) { return undefined; } if (op === 'access' || op === 'optionalAccess') { lastAccessLHS = value; value = fn(value); } else if (op === 'call' || op === 'optionalCall') { value = fn((...args) => value.call(lastAccessLHS, ...args)); lastAccessLHS = undefined; } } return value; }var _biggystring = require('biggystring');
var _cleaners = require('cleaners');











var _types = require('edge-core-js/types');










var _swapHelpers = require('../../util/swapHelpers');
var _utils = require('../../util/utils');
var _types3 = require('../types');
var _changenow = require('./changenow');

const pluginId = 'letsexchange'

const swapInfo = {
  pluginId,
  isDex: false,
  displayName: 'LetsExchange',
  supportEmail: 'support@letsexchange.io'
}

const asInitOptions = _cleaners.asObject.call(void 0, {
  apiKey: _cleaners.asString,
  affiliateId: _cleaners.asOptional.call(void 0, _cleaners.asString)
})

const orderUri = 'https://letsexchange.io/?exchangeId='
const uri = 'https://api.letsexchange.io/api/v1/'

const expirationMs = 1000 * 60

const asQuoteInfo = _cleaners.asObject.call(void 0, {
  transaction_id: _cleaners.asString,
  deposit_amount: _cleaners.asString,
  deposit: _cleaners.asString,
  deposit_extra_id: _changenow.asOptionalBlank.call(void 0, _cleaners.asString),
  withdrawal_amount: _cleaners.asString,
  withdrawal_extra_id: _changenow.asOptionalBlank.call(void 0, _cleaners.asString)
})

const asInfoReply = _cleaners.asObject.call(void 0, {
  min_amount: _types3.asNumberString,
  max_amount: _types3.asNumberString,
  amount: _types3.asNumberString
})
const INVALID_CURRENCY_CODES = {
  from: {
    ethereum: ['MATH'],
    polygon: ['USDC.e']
  },
  to: {
    ethereum: ['MATH'],
    polygon: ['USDC.e'],
    zcash: ['ZEC']
  }
}

// See https://letsexchange.io/exchange-pairs for list of supported currencies
const MAINNET_CODE_TRANSCRIPTION = {
  algorand: 'ALGO',
  arbitrum: 'ARBITRUM',
  avalanche: 'AVAXC',
  // axelar: // Shows up as WAXL from the API but is currently disabled
  // base:
  binance: 'BEP2',
  binancesmartchain: 'BEP20',
  bitcoin: 'BTC',
  bitcoincash: 'BCH',
  bitcoingold: 'BTG',
  bitcoinsv: 'BSV',
  celo: 'CELO',
  // coreum:
  cosmoshub: 'ATOM',
  dash: 'DASH',
  digibyte: 'DGB',
  dogecoin: 'DOGE',
  // eboost:
  eos: 'EOS',
  ethereum: 'ERC20',
  ethereumclassic: 'ETC',
  ethereumpow: 'ETHW',
  fantom: 'FTM',
  // feathercoin:
  filecoin: 'FIL',
  // filecoinfevm:
  fio: 'FIO',
  groestlcoin: 'GRS',
  hedera: 'HBAR',
  // liberland:
  litecoin: 'LTC',
  monero: 'XMR',
  optimism: 'OPTIMISM',
  osmosis: 'OSMO',
  piratechain: 'ARRR',
  polkadot: 'DOT',
  polygon: 'MATIC',
  pulsechain: 'PLS',
  qtum: 'QTUM',
  ravencoin: 'RVN',
  ripple: 'XRP',
  rsk: 'RSK',
  // smartcash:
  solana: 'SOL',
  stellar: 'XLM',
  telos: 'TLOS',
  tezos: 'XTZ',
  thorchainrune: 'RUNE',
  tron: 'TRC20',
  // ufo:
  // vertcoin:
  wax: 'WAX',
  zcash: 'ZEC',
  zcoin: 'FIRO',
  zksync: 'ZKSYNC'
}

const SPECIAL_MAINNET_CASES

 = {
  // axelar: new Map([[null, 'WAXL']]), // currentlyly disabled
  binancesmartchain: new Map([[null, 'BNB']]),
  ethereum: new Map([[null, 'ETH']]),
  rsk: new Map([[null, 'RBTC']]),
  tron: new Map([[null, 'TRX']])
}

const CURRENCY_CODE_TRANSCRIPTION = {
  optimism: {
    VELO: 'VELODROME'
  }
}

 function makeLetsExchangePlugin(
  opts
) {
  const { io, log } = opts
  const { fetchCors = io.fetch } = io
  const initOptions = asInitOptions(opts.initOptions)

  async function call(
    url,
    request,
    data
  ) {
    const body = JSON.stringify(data.params)

    const headers = {
      'Content-Type': 'application/json',
      Authorization: `Bearer ${initOptions.apiKey}`,
      Accept: 'application/json'
    }
    const response = await fetchCors(url, { method: 'POST', body, headers })
    if (!response.ok) {
      if (response.status === 422) {
        throw new (0, _types.SwapCurrencyError)(swapInfo, request)
      }
      throw new Error(`letsexchange returned error code ${response.status}`)
    }
    return await response.json()
  }

  const fetchSwapQuoteInner = async (
    request,
    opts
  ) => {
    const reverseQuote = request.quoteFor === 'to'

    // Grab addresses:
    const [fromAddress, toAddress] = await Promise.all([
      _utils.getAddress.call(void 0, request.fromWallet),
      _utils.getAddress.call(void 0, request.toWallet)
    ])

    // Convert the native amount to a denomination:
    const quoteAmount =
      request.quoteFor === 'from'
        ? await request.fromWallet.nativeToDenomination(
            request.nativeAmount,
            request.fromCurrencyCode
          )
        : await request.toWallet.nativeToDenomination(
            request.nativeAmount,
            request.toCurrencyCode
          )

    const { fromMainnetCode, toMainnetCode } = _swapHelpers.getCodesWithTranscription.call(void 0, 
      request,
      MAINNET_CODE_TRANSCRIPTION,
      CURRENCY_CODE_TRANSCRIPTION
    )

    const { pluginId: fromPluginId } = request.fromWallet.currencyInfo
    const networkFrom =
      _nullishCoalesce(_optionalChain([SPECIAL_MAINNET_CASES, 'access', _ => _[fromPluginId], 'optionalAccess', _2 => _2.get, 'call', _3 => _3(request.fromTokenId)]), () => (
      fromMainnetCode))

    const { pluginId: toPluginId } = request.toWallet.currencyInfo
    const networkTo =
      _nullishCoalesce(_optionalChain([SPECIAL_MAINNET_CASES, 'access', _4 => _4[toPluginId], 'optionalAccess', _5 => _5.get, 'call', _6 => _6(request.toTokenId)]), () => ( toMainnetCode))

    // Swap the currencies if we need a reverse quote:
    const quoteParams = {
      from: request.fromCurrencyCode,
      to: request.toCurrencyCode,
      network_from: networkFrom,
      network_to: networkTo,
      amount: quoteAmount
    }

    log('quoteParams:', quoteParams)

    // Calculate the amounts:
    let fromAmount, toAmount, endpoint
    if (request.quoteFor === 'from') {
      fromAmount = quoteAmount
      endpoint = 'info'
    } else {
      toAmount = quoteAmount
      endpoint = 'info-revert'
    }
    const response = await call(uri + endpoint, request, {
      params: quoteParams
    })
    const reply = asInfoReply(response)

    // Check the min/max:
    const nativeMin = reverseQuote
      ? await request.toWallet.denominationToNative(
          reply.min_amount,
          request.toCurrencyCode
        )
      : await request.fromWallet.denominationToNative(
          reply.min_amount,
          request.fromCurrencyCode
        )

    if (_biggystring.lt.call(void 0, request.nativeAmount, nativeMin)) {
      throw new (0, _types.SwapBelowLimitError)(
        swapInfo,
        nativeMin,
        reverseQuote ? 'to' : 'from'
      )
    }

    const nativeMax = reverseQuote
      ? await request.toWallet.denominationToNative(
          reply.max_amount,
          request.toCurrencyCode
        )
      : await request.fromWallet.denominationToNative(
          reply.max_amount,
          request.fromCurrencyCode
        )

    if (_biggystring.gt.call(void 0, nativeMax, '0')) {
      if (_biggystring.gt.call(void 0, request.nativeAmount, nativeMax)) {
        throw new (0, _types.SwapAboveLimitError)(
          swapInfo,
          nativeMin,
          reverseQuote ? 'to' : 'from'
        )
      }
    }

    const { promoCode } = opts
    endpoint = reverseQuote ? 'transaction-revert' : 'transaction'
    const sendReply = await call(uri + endpoint, request, {
      params: {
        deposit_amount: reverseQuote ? undefined : fromAmount,
        withdrawal_amount: reverseQuote ? toAmount : undefined,
        coin_from: request.fromCurrencyCode,
        coin_to: request.toCurrencyCode,
        network_from: networkFrom,
        network_to: networkTo,
        withdrawal: toAddress,
        return: fromAddress,
        return_extra_id: null,
        withdrawal_extra_id: null,
        affiliate_id: initOptions.affiliateId,
        promocode: promoCode != null ? promoCode : '',
        type: 'edge',
        float: false,
        isEstimate: false
      }
    })

    log('sendReply', sendReply)
    const quoteInfo = asQuoteInfo(sendReply)

    const fromNativeAmount = await request.fromWallet.denominationToNative(
      quoteInfo.deposit_amount,
      request.fromCurrencyCode
    )
    const toNativeAmount = await request.toWallet.denominationToNative(
      quoteInfo.withdrawal_amount,
      request.toCurrencyCode
    )

    // Make the transaction:
    const spendInfo = {
      tokenId: request.fromTokenId,
      spendTargets: [
        {
          nativeAmount: fromNativeAmount,
          publicAddress: quoteInfo.deposit,
          uniqueIdentifier: quoteInfo.deposit_extra_id
        }
      ],
      networkFeeOption:
        request.fromCurrencyCode.toUpperCase() === 'BTC' ? 'high' : 'standard',
      assetAction: {
        assetActionType: 'swap'
      },
      savedAction: {
        actionType: 'swap',
        swapInfo,
        orderId: quoteInfo.transaction_id,
        orderUri: orderUri + quoteInfo.transaction_id,
        isEstimate: false,
        toAsset: {
          pluginId: request.toWallet.currencyInfo.pluginId,
          tokenId: request.toTokenId,
          nativeAmount: toNativeAmount
        },
        fromAsset: {
          pluginId: request.fromWallet.currencyInfo.pluginId,
          tokenId: request.fromTokenId,
          nativeAmount: fromNativeAmount
        },
        payoutAddress: toAddress,
        payoutWalletId: request.toWallet.id,
        refundAddress: fromAddress
      }
    }

    log('spendInfo', spendInfo)

    return {
      request,
      spendInfo,
      swapInfo,
      fromNativeAmount,
      expirationDate: new Date(Date.now() + expirationMs)
    }
  }

  const out = {
    swapInfo,

    async fetchSwapQuote(
      req,
      userSettings,
      opts
    ) {
      const request = _utils.convertRequest.call(void 0, req)
      _swapHelpers.checkInvalidCodes.call(void 0, INVALID_CURRENCY_CODES, request, swapInfo)
      _swapHelpers.checkWhitelistedMainnetCodes.call(void 0, 
        MAINNET_CODE_TRANSCRIPTION,
        request,
        swapInfo
      )

      const newRequest = await _swapHelpers.getMaxSwappable.call(void 0, 
        fetchSwapQuoteInner,
        request,
        opts
      )
      const swapOrder = await fetchSwapQuoteInner(newRequest, opts)
      return await _swapHelpers.makeSwapPluginQuote.call(void 0, swapOrder)
    }
  }

  return out
} exports.makeLetsExchangePlugin = makeLetsExchangePlugin;
